import os
import numpy as np
import pandas as pd
import time
import json
import pathlib
import torch
import torch.nn as nn
import torch.optim as optim
import matplotlib.pyplot as plt
from torch.utils.data import DataLoader
from sklearn.model_selection import StratifiedShuffleSplit
from numpy import loadtxt, savetxt
from sklearn import metrics, model_selection
import math 
from train_utils import *
from models import *
from metrics import *
from train import Training 
import seaborn as sns

curr_dir_path = str(pathlib.Path().absolute())
cwd = os.getcwd() 
save_dir_path = '/logs/'
gpu_num = 0 
cuda_available = torch.cuda.is_available()
if cuda_available:
    torch.cuda.set_device(gpu_num)

manual_seed = 1
torch.manual_seed(manual_seed)
np.random.seed(manual_seed)    

def run():
    length = 80
    width = 4*length #seq of 80 nucl one hot enc
    
    cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/full_repeat_80/' 
    
    pos_seq_encoded_chr19 = loadtxt(cwd + 'chr2_5_6_9.fa_positive_80_full_repeat.csv', delimiter=',')
    print("seq length", pos_seq_encoded_chr19.shape)
    
    num_contamination = 45
    
  #  cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/chromosomes' 
 #   pos_seq_encoded_chr19 = loadtxt(cwd + '/dataset_Y_flanking.csv', delimiter=',')
  #  pos_seq_encoded_chr19 = pos_seq_encoded_chr19[:,800:1120]
  #  cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/sliding/' 
  #  pos_seq_encoded_chr19 = loadtxt(cwd + 'chr19.test_positive_spikes.csv', delimiter=',')
    
    pos_seq_encoded_chr19 = pos_seq_encoded_chr19.astype(np.float32)
    pos_seq_encoded_chr19 = torch.from_numpy(pos_seq_encoded_chr19[0:10000,:])#400:400+width])
    whole_seq = pos_seq_encoded_chr19
    

    
    ## Load model 
    curr_dir_path = str(pathlib.Path().absolute())
    save_dir_path = 'logs/'    
    
    ################### SLIDING ##############
    json_dir = "model0_CNN_2d_FILTER1_6.pt" #save_dir_path + "model0.pt"
    with open(curr_dir_path + "/config.json", encoding='utf-8', errors='ignore') as json_data:
            config = json.load(json_data, strict=False)


    #load trained model 
    device =  torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    model = CNN_2d(device)
    model.load_state_dict(torch.load(json_dir, map_location=torch.device('cpu')))
    """
    ###################80bp ##############
    json_dir = "model52.pt" #save_dir_path + "model0.pt"
    with open(curr_dir_path + "/config_2.json", encoding='utf-8', errors='ignore') as json_data:
            config = json.load(json_data, strict=False)
    
    
    #load trained model 
    device =  torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    model = CNN_hp_tuning(device, no_conv=config['MODEL']['Conv no'], no_filters=config['MODEL']['Filter no'], 
                                       filter_size=config['MODEL']['Filter size'], stride=config['MODEL']['Conv stride'], 
                                       pool_size=config['MODEL']['Pool size'], pool_stride=config['MODEL']['Pool stride'],
                                       dense_output=config['MODEL']['FC layer output'], no_dense_layers = config['MODEL']['No FC layers'],
                                       dropout=config['MODEL']['dropout'])
    
    model.load_state_dict(torch.load(json_dir, map_location=torch.device('cpu')))
    """
    obj = Training(config, save_dir_path)
    obj.model = model
    
    
    spikes = np.zeros(whole_seq.shape[0])
    noise = np.zeros(num_contamination)
    noise_only_found = np.zeros(num_contamination)
    ##track the repeats that the model finds initially (with no noise)
    sequences = whole_seq

    for i in range(0, whole_seq.shape[0]):
            
        seq = sequences[i] 
        seq = seq.reshape(-1, length, 4)
        seq = np.transpose(seq, (0, 2, 1))
        obj.model.eval()
        output = obj.model.forward(seq)  
        spikes[i] = output
        
    repeat_indexes = spikes > 0.5        
    
    seq_noise = whole_seq[repeat_indexes]
    spikes = np.zeros(seq_noise.shape[0])
    print("identified num of repeats", seq_noise.shape[0])
    position_change = np.random.randint(length, size=num_contamination)
    print(position_change)
    seq_noise = seq_noise.reshape(-1, length, 4)
    seq_noise = np.transpose(seq_noise, (0, 2, 1))
    
    num_repeats_found = np.zeros(num_contamination)
    for y in range(0,num_contamination): 
        
        indices = np.arange(4)
        np.random.shuffle(indices) 
        

       # print(indices)
        change = seq_noise[:,indices,position_change[y]]
        seq_noise[:,:,position_change[y]] = change
        
        for i in range(0, seq_noise.shape[0]):
            
            seq = seq_noise[i] 
            seq = seq.unsqueeze(0)
            obj.model.eval()
            output = obj.model.forward(seq)  
            spikes[i] = output
        
        temp = spikes.copy()
        repeat_indexes = temp > 0.5
        print(spikes.shape)
        print(spikes)
        print(temp.shape)
        print("iteration", y)
        num_repeats_found[y] = seq_noise[repeat_indexes].shape[0]
        print("repeats identified", num_repeats_found[y])
      #  plt.plot(range(spikes.shape[0]), spikes)
        
      #  plt.plot(range(4*length), np.reshape(seq, (-1, 4*length))[0,:])
    
      #  plt.show()
        sum_ = np.sum(spikes)/seq_noise.shape[0]
        sum_only_found = np.sum(spikes[repeat_indexes])/num_repeats_found[y]
       # print(sum_)
        noise[y]= sum_
        noise_only_found[y] = sum_only_found
        
    print(noise)
    print("noise_only_found", noise_only_found)
    print("Num of repeats found", num_repeats_found)
    plt.plot(range(noise.shape[0]), noise)
    
   # plt.show()

if __name__ == "__main__":

    run()