import torch
import argparse
import time
import torch
import torch.utils.data
import torch.optim as optim
import numpy as np
import math
import random
import torch.utils.data as data_utils

from sklearn.model_selection import StratifiedShuffleSplit
from sklearn import metrics, model_selection

#import models.VAE as VAE
#from models.model import MLP_encoder, MLP_decoder, CNN_encoder, CNN_decoder
#from optimization.training import train, evaluate, latent_repr
from load_data import load_dataset
import json
import pathlib
import time
from numpy import loadtxt, savetxt
import matplotlib.pyplot as plt

import numpy as np 
import pandas as pd
from pandas.plotting import scatter_matrix
import seaborn as sns
from vae_gumbel_softmax import VAE_gumbel, VAE_gumbel_hp_tuning, latent_repr, evaluate
from cnn.train_multi import Training

parser = argparse.ArgumentParser(description='PyTorch Normalizing flows')

parser.add_argument('--incl_rec', type=bool, default=True, metavar='incl_rec',
                    help='include reconstruction error')

parser.add_argument('--only_rec', type=bool, default=True, metavar='only_rec',
                    help='include ONLY reconstruction error')
parser.add_argument('--data', type=str, default='before', metavar='data',
                    help='training data type')

parser.add_argument('-nc', '--no_cuda', action='store_true', default=False,
                    help='disables CUDA training')

parser.add_argument('--manual_seed', type=int, help='manual seed, if not given resorts to random seed.')

parser.add_argument('-li', '--log_interval', type=int, default=10, metavar='LOG_INTERVAL',
                    help='how many batches to wait before logging training status')

parser.add_argument('-od', '--out_dir', type=str, default='logs/', metavar='OUT_DIR',
                    help='output directory for model snapshots etc.')

fp = parser.add_mutually_exclusive_group(required=False)
fp.add_argument('-te', '--testing', action='store_true', dest='testing',
                help='evaluate on test set after training')
fp.add_argument('-va', '--validation', action='store_false', dest='testing',
                help='only evaluate on validation set')
parser.set_defaults(testing=False)


parser.add_argument('-bs', '--batch_size', type=int, default=256, metavar='BATCH_SIZE',
                    help='input batch size for training (default: 100)')

parser.add_argument('--z_size', type=int, default=40, metavar='ZSIZE',
                    help='how many stochastic hidden units')
parser.add_argument('--flat_seq_size', type=int, default=480, metavar='FSSIZE',
                    help='size of flattened sequence')
parser.add_argument('--encoder_dim', type=int, default=40, metavar='ESIZE',
                    help='output feature dim of encoder')
parser.add_argument('--decoder_dim', type=int, default=480, metavar='DSIZE',
                    help='output feature dim of decoder') 

parser.add_argument('--cnn_enc', type=bool, default=True, metavar='cnn_enc',
                    help='encoder cnn')
parser.add_argument('--cnn_dec', type=bool, default=True, metavar='cnn_dec',
                    help='decoder cnn')
parser.add_argument('--test_model', type=bool, default=True, metavar='test_model',
                    help='test model')
# gpu/cpu
parser.add_argument('--gpu_num', type=int, default=0, metavar='GPU', help='choose GPU to run on.')



args = parser.parse_args()
args.cuda = not args.no_cuda and torch.cuda.is_available()

if args.manual_seed is None:
    args.manual_seed = random.randint(1, 100000)
random.seed(args.manual_seed)
torch.manual_seed(args.manual_seed)
np.random.seed(args.manual_seed)

if args.cuda:
    torch.cuda.set_device(args.gpu_num)


curr_dir_path = str(pathlib.Path().absolute()) 
save_dir_path = '/cnn/' + '/logs/'
gpu_num = 0 
cuda_available = torch.cuda.is_available()
if cuda_available:
    torch.cuda.set_device(gpu_num)
    
def run(args):
    
    ##################################### Load data #####################################
    if args.data == 'before':
        print("data type: sequences before the repeat")
        length = 40
        cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/before_data/'        
        neg_seq_encoded_total = loadtxt(cwd + 'chr19.test_negative_before.csv', delimiter=',')
        neg_seq_encoded_total2 = loadtxt(cwd + 'chr19.test_negative_before_2.csv', delimiter=',')
        pos_seq_encoded_chr19 = loadtxt(cwd + 'chr19.test_positive_before.csv', delimiter=',')
        pos_seq_encoded_chr21 = loadtxt(cwd + 'chr21.test_positive_before.csv', delimiter=',')
        pos_seq_encoded_chr22 = loadtxt(cwd + 'chr22.test_positive_before.csv', delimiter=',')
        pos_seq_encoded_chr7 = loadtxt(cwd + 'chr7.test_positive_before.csv', delimiter=',')
        pos_seq_encoded_chr8 = loadtxt(cwd + 'chr8.test_positive_before.csv', delimiter=',')
        
        pos_seq_encoded_total = np.concatenate((pos_seq_encoded_chr19, pos_seq_encoded_chr21, pos_seq_encoded_chr22, pos_seq_encoded_chr7, pos_seq_encoded_chr8), axis = 0)   
        neg_seq_encoded_total = np.append(neg_seq_encoded_total, neg_seq_encoded_total2, axis = 0)
        
    elif args.data == '80bp':
        length = 80
        print("data type: 80bp containg the repeat - initial position between 1-20")
        cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/window_80/'       
        pos_seq_encoded_chr19 = loadtxt(cwd + 'chr19.fa_positive_80.csv', delimiter=',')  
        pos_seq_encoded_chr21 = loadtxt(cwd + 'chr21.fa_positive_80.csv', delimiter=',')
        pos_seq_encoded_chr22 = loadtxt(cwd + 'chr22.fa_positive_80.csv', delimiter=',')
        pos_seq_encoded_chr7 = loadtxt(cwd + 'chr7.fa_positive_80.csv', delimiter=',')
        neg_seq_encoded_chr19 = loadtxt(cwd + 'chr19.fa_negative_80_uscs.csv', delimiter=',')
        pos_seq_encoded_total = np.concatenate((pos_seq_encoded_chr19, pos_seq_encoded_chr21, pos_seq_encoded_chr22, pos_seq_encoded_chr7), axis = 0)
        neg_seq_encoded_total =  neg_seq_encoded_chr19 #np.append(neg_seq_encoded_chr19, neg_seq_encoded_chr7, axis = 0)
        
    elif args.data == '60bp':
        length = 60
        print("data type: 60bp containg the repeat - initial position between 1-20")
        cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/window_60/'        
        #pos_seq_encoded_total = loadtxt(cwd + 'chr19_21_22_7_8.fa_positive_60.csv', delimiter=',')  
        neg_seq_encoded_total = loadtxt(cwd + 'chr19.fa_negative_60_uscs.csv', delimiter=',')
        cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/window_60_pos_0/'        
        pos_seq_encoded_total = loadtxt(cwd + 'chr19_21_22_7.fa_positive_60_pos0.csv', delimiter=',')  
        
    elif args.data == 'sliding':
        length = 80
        print("data type: sliding windows, length 80")
        cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/sliding/'       
        pos_seq_encoded_chr19 = loadtxt(cwd + 'chr19.test_positive_spikes.csv', delimiter=',')  
        pos_seq_encoded_chr19_2 = loadtxt(cwd + 'chr19.test_positive_spikes2.csv', delimiter=',')
        pos_seq_encoded_total = np.concatenate((pos_seq_encoded_chr19, pos_seq_encoded_chr19_2), axis = 0)
        neg_seq_encoded_chr19_1 = loadtxt(cwd + 'chr19.test_negative_spikes_uscs.csv', delimiter=',')
        neg_seq_encoded_chr19_2 = loadtxt(cwd + 'chr19.test_negative_spikes2_uscs.csv', delimiter=',')
        neg_seq_encoded_chr19_3 = np.zeros((5000, neg_seq_encoded_chr19_1.shape[1]))
        neg_seq_encoded_total = np.concatenate((neg_seq_encoded_chr19_1, neg_seq_encoded_chr19_2, neg_seq_encoded_chr19_3), axis = 0) 
        
    elif args.data == 'sliding_60':
        length = 60
        print("data type: sliding windows, length 60")
        cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/sliding_60/'       
        pos_seq_encoded_total = loadtxt(cwd + 'chr19.fa_positive_sliding_60.csv', delimiter=',')  
        pos_seq_encoded_4 = loadtxt(cwd + 'chr19.fa_positive_60_sliding_4.csv', delimiter=',')  
        pos_seq_encoded_5 = loadtxt(cwd + 'chr19.fa_positive_60_sliding_5.csv', delimiter=',')  
        pos_seq_encoded_total = np.concatenate((pos_seq_encoded_total, pos_seq_encoded_4, pos_seq_encoded_5), axis = 0)
        cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/sliding/'   
        neg_seq_encoded_chr19_1 = loadtxt(cwd + 'chr19.test_negative_spikes_uscs.csv', delimiter=',')
        
        neg_seq_encoded_chr19_2 = loadtxt(cwd + 'chr19.test_negative_spikes2_uscs.csv', delimiter=',')
        neg_seq_encoded_chr19_3 = np.zeros((5000, neg_seq_encoded_chr19_1.shape[1]))
        neg_seq_encoded_total = np.concatenate((neg_seq_encoded_chr19_1, neg_seq_encoded_chr19_2, neg_seq_encoded_chr19_3), axis = 0) 
        
        neg_seq_encoded_total = neg_seq_encoded_total[:,0:length*4]
        
    elif args.data == 'starting_position':
        length = 60
        print("data type: different starting positions of the repeats")
        cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/window_60/' 
        pos_seq_encoded_total =  loadtxt(cwd + 'chr19_21_22_7_8.fa_positive_60.csv', delimiter=',') 
        neg_seq_encoded_total = loadtxt(cwd + 'chr19.fa_negative_60_uscs.csv', delimiter=',')
        cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/starting_position/' 
      #  pos_seq_encoded_total_repeats = loadtxt(cwd + 'chr22.fa_positive_no_up_down_120.csv', delimiter=',')
        pos_seq_encoded_total_repeats_pos20 = loadtxt(cwd + 'chr21.fa_test_positive_80_spikes_stpos20.csv', delimiter=',')
        
        pos_seq_encoded_total_repeats = pos_seq_encoded_total_repeats_pos20[:,400:400 + 4*length] #starting point of repeat-pos0
        pos_seq_encoded_total_repeats_pos10 = pos_seq_encoded_total_repeats_pos20[:,360:360 + 4*length] #starting point of repeat-pos10

        pos_seq_encoded_total_repeats_pos30 = pos_seq_encoded_total_repeats_pos20[:,280:280 + 4*length] #starting point of repeat-pos30
        pos_seq_encoded_total_before = pos_seq_encoded_total_repeats_pos20[:,160:400] #seq just before the repeat
        
        #select which will be the positive and negative classes
        pos_seq_encoded_total = pos_seq_encoded_total_repeats
        neg_seq_encoded_total = pos_seq_encoded_total_repeats_pos10
    elif args.data == 'full_repeat':
        length = 60
        cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/full_repeat_60/'       
        pos_seq_encoded_total = loadtxt(cwd +'chr19_21_22_7_8.fa_positive_60_full_repeat.csv', delimiter=',')  
        
        cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/window_60/' 
        
        neg_seq_encoded_total = loadtxt(cwd +'chr19.fa_negative_60_uscs.csv', delimiter=',')
        
            
    elif args.data == 'before_60bp':
        length = 60
        # positive data: part just before the repeat starts
        #negative data: part where the is not a repeat for at least 
        cwd = 'C:/Users/30697/Desktop/Deep-Learning-DNA-Sequences-master/Code/split/correct_split/VAE_gumble_softmax-master/data/before_data_60/' 
      #  pos_seq_encoded_total_repeats = loadtxt(cwd + 'chr22.fa_positive_no_up_down_120.csv', delimiter=',')
        pos_seq_encoded_total_repeats_pos20 = loadtxt(cwd + 'chr19.fa_positive_60_before.csv', delimiter=',')
        pos_seq_encoded_total_before = pos_seq_encoded_total_repeats_pos20 ##seq just before the repeat
        pos_seq_encoded_total = pos_seq_encoded_total_before
        neg_seq_encoded_total_before  = loadtxt(cwd + 'chr19.fa_negative_60_before1_uscs.csv', delimiter=',')
        print("neg data shape", neg_seq_encoded_total_before.shape)
        neg_seq_encoded_total = neg_seq_encoded_total_before
        
    ###################################################################################
    print(args.batch_size)
    
    N = neg_seq_encoded_total.shape[0]
    indices = np.arange(N)
    np.random.shuffle(indices)
    neg_seq_encoded_total = neg_seq_encoded_total[indices, :]
    
    N = pos_seq_encoded_total.shape[0]
    indices = np.arange(N)
    np.random.shuffle(indices)
    pos_seq_encoded_total = pos_seq_encoded_total[indices, :]
    
    end = min(args.batch_size*int(pos_seq_encoded_total.shape[0]/args.batch_size), args.batch_size*int(neg_seq_encoded_total.shape[0]/args.batch_size))
    
    neg_seq_encoded_total = neg_seq_encoded_total[0:end]  
    pos_seq_encoded_total = pos_seq_encoded_total[0:end]
      
    print("negative", neg_seq_encoded_total.shape)
    print("positive", pos_seq_encoded_total.shape)
    
    y_pos = np.ones((pos_seq_encoded_total.shape[0], 1))
    y_neg = np.zeros((neg_seq_encoded_total.shape[0], 1))
    
    sequences = np.concatenate((pos_seq_encoded_total, neg_seq_encoded_total), axis=0)
    labels = np.concatenate((y_pos, y_neg), axis=0)
    print("positives", y_pos.shape)
    print("negatives", y_neg.shape)
    
    #Shuffle the data (positive and negative examples should be shuffled)
    N = labels.shape[0]
    indices = np.arange(N)
    np.random.shuffle(indices)
    sequences = sequences[indices, :]
    labels = labels[indices]
    print("total dataset size", sequences.shape)
    
    data2, args = load_dataset(args, sequences)
    
    # Load trained VAE model 
 #   model = VAE_gumbel_hp_tuning(0.7, 4, 3, 25, 25, 4, 1, 2, 2, length, 'sigmoid')
 #   json_dir = "model_VAE.pt"
    
    model = VAE_gumbel_hp_tuning(0.7, 4, 3, 25, 25, 4, 1, 2, 2, length, 'sigmoid')
    json_dir = "model0_7.pt"
    
    model.load_state_dict(torch.load(json_dir, map_location=torch.device('cpu')))    
    print(model)
    if args.cuda:
        model.cuda()
        print("Model on GPU")
        
    latent2 = latent_repr(data2, model, args)
    
    dim = latent2.shape[1]
    print("dimension", dim) 
    if args.only_rec:
        print("including ONLY the rec error")
        latent2 = 0*latent2
    if args.incl_rec ==1:
        print("Including the reconstruction error as an extra feature")
        loss2 = evaluate(data2, model, args) 
        loss2 = np.expand_dims(loss2, 1)
        sequences_latent = np.concatenate((latent2, loss2), axis=1)
        dim += 1
    else:    
        sequences_latent = latent2
    
    print("shape sequences for cnn and mlp", sequences.shape, sequences_latent.shape)
    ### sequences_latent -- inputs for the MLP 
    ### sequences -- inputs for the feed-forward
    sequences_ = np.concatenate((sequences, sequences_latent), axis=1)
    print("shape after concatenation", sequences_.shape)
    
    # Stratified splits 
    train_val_data, test_data, train_val_labels, test_labels = model_selection.train_test_split(sequences_, labels, test_size=0.05,
                                                                                                        random_state=1, stratify=labels)
    train_data, val_data, train_labels, val_labels = model_selection.train_test_split(train_val_data,train_val_labels,test_size=0.05,
                                                                                                  random_state=1, stratify=train_val_labels)
    
    
    train_accuracy_curve = []
    train_f1_curve = []
    val_accuracy_curve = []
    val_f1_curve = []
    
    #train the model 
    print("start training the model in the latent space")
    # Get config file
    with open(curr_dir_path + "/cnn" + "/config.json", encoding='utf-8', errors='ignore') as json_data:
        config = json.load(json_data, strict=False)   

    config['TRAINER']['save_dir'] = save_dir_path
    obj = Training(config, train_data, train_labels, val_data, val_labels, test_data, test_labels, dim, save_dir_path, input_length=length)
    train_accuracy_curve, train_f1_curve, val_accuracy_curve, val_f1_curve = obj.training_pipeline()
     
    #Plot training and validation curves (accuracy, f1)
    plt.plot(train_accuracy_curve, label='train_accuracy')
    plt.plot(val_accuracy_curve, label = 'val_accuracy')
    plt.plot(train_f1_curve, label='train_f1')
    plt.plot(val_f1_curve, label = 'val_f1')
    plt.xlabel('Epoch')
    plt.ylabel('Accuracy')
    plt.ylim([0.5, 1.0])
    plt.legend(loc='lower right') 
    
if __name__ == "__main__":

    run(args)