import torch
import torch.utils.data as data_utils
import pickle
from scipy.io import loadmat
from numpy import loadtxt, savetxt


import os
import numpy as np
import pandas as pd
import time
import json
import pathlib
import torch
import torch.nn as nn
import torch.optim as optim
import matplotlib.pyplot as plt
from torch.utils.data import DataLoader
from sklearn.model_selection import StratifiedShuffleSplit
from sklearn import metrics, model_selection

import numpy as np

import os


def load_mnist(args, data, VAE_training):
    """
    Dataloading function for static mnist. Outputs image data in vectorized form: each image is a vector of size 784
    """
 #   args.input_size = [4* 80]
 #   args.input_dim = 4*80
    
    if args.test_model == 0:
        x_train, y_train, x_val, y_val, x_test, y_test = load_data(data, args, VAE_training)

        # pytorch data loader
        train = data_utils.TensorDataset(torch.from_numpy(x_train).float(), torch.from_numpy(y_train))
        train_loader = data_utils.DataLoader(train, batch_size=args.batch_size, shuffle=True)

        validation = data_utils.TensorDataset(torch.from_numpy(x_val).float(), torch.from_numpy(y_val))
        val_loader = data_utils.DataLoader(validation, batch_size=args.batch_size, shuffle=True)

        test = data_utils.TensorDataset(torch.from_numpy(x_test).float(), torch.from_numpy(y_test))
        test_loader = data_utils.DataLoader(test, batch_size=args.batch_size, shuffle=True)

        return train_loader, val_loader, test_loader, args
    
    else:
        data_train = load_data(data, args, VAE_training)
        y_train = np.zeros((data_train.shape[0], 1))
        train = data_utils.TensorDataset(torch.from_numpy(data_train).float(), torch.from_numpy(y_train))
        train_loader = data_utils.DataLoader(train, batch_size=args.batch_size, shuffle=False)
        return train_loader, args

    

def load_dataset(args, data, VAE_training=False):

    if args.test_model == 0:
        train_loader, val_loader, test_loader, args = load_mnist(args, data, VAE_training)
        return train_loader, val_loader, test_loader, args
    else:
        data_loader, args = load_mnist(args, data, VAE_training)
        return data_loader, args
    

def load_data(data, args, VAE_training):
            args.cnn_enc = 1            
            seq_encoded = data 
            y_label = np.ones(seq_encoded.shape[0])   #this label is unused [TODO]

            #Shuffle the data (positive and negative examples should be shuffled)
            N = y_label.shape[0]
            indices = np.arange(N)
            np.random.shuffle(indices)
            seq_encoded = seq_encoded[indices, :]
            y_label = y_label[indices]
                        
            if args.test_model == 1:               
                if (args.cnn_enc):
                    seq_encoded = seq_encoded.reshape(-1, args.length, 4)
                    seq_encoded = np.transpose(seq_encoded, (0, 2, 1))
                    seq_encoded = seq_encoded.astype(np.float32)
                    args.input_size = [4, args.length]
                    args.input_dim = [4, args.length]

                return seq_encoded
            
            
            
            # Split (train+val) and test  
            # Stratified splits 
            if VAE_training==True:
                val_size = 0.00001
            else:
                val_size = 0.05
                
            train_val_data, test_data, train_val_labels, test_labels = model_selection.train_test_split(seq_encoded, y_label, test_size=0.01, random_state=1)
            train_data, val_data, train_labels, val_labels = model_selection.train_test_split(train_val_data, train_val_labels, test_size=val_size, random_state=1)


            if (args.cnn_enc):
            # Reshape training image
                train_data = train_data.reshape(-1, args.length, 4)
                val_data = val_data.reshape(-1, args.length, 4)
                test_data = test_data.reshape(-1, args.length, 4)
                train_data = np.transpose(train_data, (0, 2, 1))
                val_data = np.transpose(val_data, (0, 2, 1))
                test_data = np.transpose(test_data, (0, 2, 1))
                args.input_size = [4, args.length]
                args.input_dim = [4, args.length]
             
                            
            train_data = train_data.astype(np.float32)
            val_data = val_data.astype(np.float32)
            test_data = test_data.astype(np.float32)
            
            print('Training data shape : ', train_data.shape, train_labels.shape)
            print('Validation data shape : ', val_data.shape, val_labels.shape)
            print('Testing data shape : ', test_data.shape, test_labels.shape)
            print(train_data.dtype)
            print(val_data.dtype)
            print(test_data.dtype)

            return train_data, train_labels, val_data, val_labels, test_data, test_labels
        
def random_shuffle(arr):
    x, y, z = arr.shape
    rows = np.indices((x,y))[0]
    cols = [np.random.permutation(y) for _ in range(x)]
    return arr[rows, cols, :]  
