import torch
import argparse
import time
import torch
import torch.utils.data
import torch.optim as optim
import numpy as np
import math
import random
import torch.utils.data as data_utils

from load_data import load_dataset
import json
import pathlib
import time
from numpy import loadtxt, savetxt
import matplotlib.pyplot as plt

import numpy as np 
import pandas as pd
from pandas.plotting import scatter_matrix
import seaborn as sns
from encode import decode_seq, one_hot_to_nucl
from vae_gumbel_softmax import VAE_gumbel, reconstruction, VAE_gumbel_hp_tuning

parser = argparse.ArgumentParser(description='PyTorch Normalizing flows')

parser.add_argument('-d', '--dataset', type=str, default='mnist', choices=['mnist'],
                    metavar='DATASET',
                    help='Dataset choice.')

parser.add_argument('-nc', '--no_cuda', action='store_true', default=False,
                    help='disables CUDA training')

parser.add_argument('--manual_seed', type=int, help='manual seed, if not given resorts to random seed.')

parser.add_argument('-li', '--log_interval', type=int, default=10, metavar='LOG_INTERVAL',
                    help='how many batches to wait before logging training status')

parser.add_argument('-od', '--out_dir', type=str, default='logs/', metavar='OUT_DIR',
                    help='output directory for model snapshots etc.')

fp = parser.add_mutually_exclusive_group(required=False)
fp.add_argument('-te', '--testing', action='store_true', dest='testing',
                help='evaluate on test set after training')
fp.add_argument('-va', '--validation', action='store_false', dest='testing',
                help='only evaluate on validation set')
parser.set_defaults(testing=False)

# optimization settings
parser.add_argument('-e', '--epochs', type=int, default=1, metavar='EPOCHS',
                    help='number of epochs to train (default: 1000)')
parser.add_argument('-es', '--early_stopping_epochs', type=int, default=100, metavar='EARLY_STOPPING',
                    help='number of early stopping epochs')

parser.add_argument('-bs', '--batch_size', type=int, default=1, metavar='BATCH_SIZE',
                    help='input batch size for training (default: 100)')
parser.add_argument('-lr', '--learning_rate', type=float, default=0.00001, metavar='LEARNING_RATE',
                    help='learning rate')

parser.add_argument('-a', '--anneal', type=str, default="kl", choices= ["std", "off", "kl"], help="beta annealing scheme")
parser.add_argument('--max_beta', type=float, default=1., metavar='MB',
                    help='max beta for warm-up')
parser.add_argument('--min_beta', type=float, default=0.0, metavar='MB',
                    help='min beta for warm-up')
parser.add_argument('-f', '--flow', type=str, default='planar', choices=['planar', 'NICE', 'NICE_MLP', 'syl_orthogonal', 'real' ])
parser.add_argument('-nf', '--num_flows', type=int, default=0,
                    metavar='NUM_FLOWS', help='Number of flow layers, ignored in absence of flows')
parser.add_argument('-nov', '--num_orthonormal_vec', type=int, default=8, metavar='NUM_ORTHONORMAL_VEC',
                    help='For orthogonal flow: number of orthogonal vectors (M)')
parser.add_argument('--syl_ortho_invertible', type=bool, default=False, metavar='ORTHO_FLOW_INVERT',
                    help='select if we need the sylvester orthogonal flow to be invertible')
parser.add_argument('--z_size', type=int, default=40, metavar='ZSIZE',
                    help='how many stochastic hidden units')
parser.add_argument('--flat_seq_size', type=int, default=320, metavar='FSSIZE',
                    help='size of flattened sequence')
parser.add_argument('--encoder_dim', type=int, default=40, metavar='ESIZE',
                    help='output feature dim of encoder')
parser.add_argument('--decoder_dim', type=int, default=480, metavar='DSIZE',
                    help='output feature dim of decoder') 

parser.add_argument('--cnn_enc', type=bool, default=True, metavar='cnn_enc',
                    help='encoder cnn')

parser.add_argument('--cnn_dec', type=bool, default=True, metavar='cnn_dec',
                    help='decoder cnn')
parser.add_argument('--test_model', type=bool, default=True, metavar='test_model',
                    help='test model')
# gpu/cpu
parser.add_argument('--gpu_num', type=int, default=0, metavar='GPU', help='choose GPU to run on.')
parser.add_argument('--temp', type=float, default=1.0, metavar='S',
                    help='tau(temperature) (default: 1.0)')
args = parser.parse_args()
args.cuda = not args.no_cuda and torch.cuda.is_available()

if args.manual_seed is None:
    args.manual_seed = random.randint(1, 100000)
random.seed(args.manual_seed)
torch.manual_seed(args.manual_seed)
np.random.seed(args.manual_seed)

if args.cuda:
    torch.cuda.set_device(args.gpu_num)

    
def run(args):
    
    pos_seq_encoded_chr19 = loadtxt('chr19.fa_positive_80.csv', delimiter=',')
    data1, args = load_dataset(args, pos_seq_encoded_chr19[4:5])
  #  print(pos_seq_encoded_chr19.shape)
 #   pos_seq_encoded_chr7 = loadtxt('chr7.fa_positive_120.csv', delimiter=',')
 #   data2, args = load_dataset(args, pos_seq_encoded_chr7[0:1000])
   # plt.plot(range(args.flat_seq_size), pos_seq_encoded_chr19[0])
  #  neg_seq_encoded_chr19 = loadtxt('neg_seq_encoded_chr21_120_.csv', delimiter=',')
  #  data1, args = load_dataset(args, neg_seq_encoded_chr19[0:10])
    
    #see dataloader 
    
    
 #   model = VAE_gumbel(args.temp)
    model = VAE_gumbel_hp_tuning(0.7, 3, 2, 12, 8, 5, 1, 2, 2)
    json_dir = "model_server.pt"
    model.load_state_dict(torch.load(json_dir, map_location=torch.device('cpu')))
 
 
    rec1, orig1, q, z, decoded = reconstruction(data1, model)
    print("q", q)
    print("z", z)
    print(z.type)
    print("decoded", decoded)
    rec1 = np.transpose(rec1, (0, 2, 1))
    orig1 = np.transpose(orig1, (0, 2, 1))
    orig1 = orig1.reshape(-1, 80*4)
    rec1 = rec1.reshape(-1, 80*4)
 #   print(rec1[0])
 #   print(orig1[0])
 #   reconstructed_seq = decode_seq(rec1[0])
    original_seq = decode_seq(orig1[0])
 #   print(reconstructed_seq)
 #   print(original_seq)
    
   # rec_one_hot = one_hot_to_nucl(reconstructed_seq)
 #   print(orig1[0].shape)
  #  plt.plot(range(args.flat_seq_size), orig1[0])
  #  plt.plot(range(args.flat_seq_size), rec1[0])
    plt.plot(range(args.flat_seq_size), orig1[0] - one_hot_to_nucl(rec1[0]))
    plt.show()
 #   rec2 = reconstruction(data2, model, args)
 #   rec3 = reconstruction(data3, model, args)
    
 #   print("shape before ", rec1.shape)
 #   x = np.ones((latent1.shape[0], 1))
 #   latent1 = np.concatenate((latent1, x), axis=1)
 #   print("shape after ", latent1.shape)
 #   latent2 = np.concatenate((latent2, 2*x), axis=1)
 #   latent3 = np.concatenate((latent3, 3*x), axis=1)
    
  #  data = np.concatenate((latent1, latent2, latent3), axis=0)
  #  df1 = pd.DataFrame(data, columns=['1','2','3','4','5','6','7','8','9','10','11'])
#    df2 = pd.DataFrame(latent2, columns=['1','2','3','4','5','6','7','8','9','10'])
#    df3 = pd.DataFrame(latent3, columns=['1','2','3','4','5','6','7','8','9','10'])
    
  #  print(latent1.shape)
  #  print(latent1[:,0])
    #plot 2 dim of latent space
    
   # sns.pairplot(df1, hue='11')
    #pd.plotting.scatter_matrix(df1)
    #pd.plotting.scatter_matrix(df2)
    #pd.plotting.scatter_matrix(df3)

  #  plt.scatter(latent1[:,0], latent1[:,1])
  #  plt.scatter(latent2[:,0], latent2[:,1])
  #  plt.scatter(latent3[:,0], latent3[:,1])
  #  plt.show()
    x = orig1[0]
    x = np.reshape(x, (80, 4))
    x = np.transpose(x, (1, 0))
    ax = plt.subplot(2, 1, 1)
    ax.imshow(x)
    
    x = rec1[0]
    x = np.reshape(x, (80, 4))
    x = np.transpose(x, (1, 0))
    ax = plt.subplot(2, 1, 2)
    ax.imshow(x)
if __name__ == "__main__":
    

    run(args)