import argparse
import os
import sys
import math
import random 
### Spliting the data according to our different types of datasets

def isfile(path):
    """Check if path is an existing file.
      Arguments:
          path: Path to the file
    """
    if not os.path.isfile(path):
        if os.path.isdir(path):
            msg = "{0} is a directory".format(path)
        else:
            msg = "{0} does not exist.".format(path)
        raise argparse.ArgumentTypeError(msg)
    return path


def isdir(path):
    """Check if path is an existing file.
      Arguments:
          path: Path to the file
    """
    if not os.path.isdir(path):
        if os.path.isfile(path):
            msg = "{0} is a file.".format(path)
        else:
            msg = "{0} does not exist.".format(path)
        raise argparse.ArgumentTypeError(msg)
    return path

def get_parameters():
    parser = argparse.ArgumentParser(description=__doc__,
            formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-i', '--input-file', dest='fasta_file', required=True,
            default=argparse.SUPPRESS, type=isfile,
            help='Input FASTA file to split')
    parser.add_argument('-o', '--output-dir', dest='output_dir', type=isdir,
            default=os.curdir + os.sep, help='Output directory')    
    parser.add_argument('-s', '--spikes', type=int, default=1, metavar='spikes',
                    help='determines the way the data are splitted')
    
    return parser.parse_args(), parser
    

def read_fasta(fasta_stream):
    header, seq = None, []
    
    for line in fasta_stream:
        line = line.rstrip()
        if line.startswith(">"):
            if header:
                yield (header, ''.join(seq))
            header, seq = line, []
        else:
            seq.append(line)
    
    if header:
        yield (header, ''.join(seq))
    

def split(fasta_file, spikes, output_dir):
    """
    Splits text in lines of specific width and then 
    creates the training set (positive and negatives examples are split in two different txt files)
    """
    (fasta_file_basename,
            fasta_file_extension)  = os.path.splitext(os.path.basename(fasta_file))
    output_file_prefix = os.path.join(output_dir, fasta_file_basename)
    try:
        with open(fasta_file, 'rt') as stream:
            
            
            
            #Split text in lines of length specific width (=fixed window) and
            #separate negative from positive training example
            for header, seq in read_fasta(stream):
                if spikes==0:
                    new_file = "{0}_{1}{2}".format(output_file_prefix, "negative_80bp", ".txt")
                    file_neg = open(new_file, 'wt')
                    new_file = "{0}_{1}{2}".format(output_file_prefix, "positive_80bp", ".txt")
                    file_pos = open(new_file, 'wt')
                    negatives, positives = fill(seq)
                elif spikes==1:
                    new_file = "{0}_{1}{2}".format(output_file_prefix, "negative", ".txt")
                    file_neg = open(new_file, 'wt')
                    new_file = "{0}_{1}{2}".format(output_file_prefix, "positive", ".txt")
                    file_pos = open(new_file, 'wt')
                    negatives, positives = fill_spikes(seq)
                    print("spliting file: spikes mode")
                
                elif spikes==2:
                    new_file = "{0}_{1}{2}".format(output_file_prefix, "negative_80_before_gap_10", ".txt")
                    file_neg = open(new_file, 'wt')
                    new_file = "{0}_{1}{2}".format(output_file_prefix, "positive_80_before_gap_10", ".txt")
                    file_pos = open(new_file, 'wt')
                    negatives, positives = fill_before(seq)
                    print("spliting file: before mode")
                    
                elif spikes==3:
                    new_file = "{0}_{1}{2}".format(output_file_prefix, "negative_80_before", ".txt")
                    file_neg = open(new_file, 'wt')
                    new_file = "{0}_{1}{2}".format(output_file_prefix, "positive_80_before", ".txt")
                    file_pos = open(new_file, 'wt')
                    negatives, positives = fill_before_spikes(seq)
                    print("spliting file: before_spikes mode")
                elif spikes==4:
                    new_file = "{0}{1}".format("deaseas_loci_14", ".txt")
                    file_dis = open(new_file, 'wt')
                    disease_repeat = fill_disease(seq)
                    print("spliting file: disease repeats mode")
                
                if spikes==4:
                    sys.stdout = file_dis # Change the standard output to the file we created.
                    print('{1}{0}{2}'.format(os.linesep, header, disease_repeat))
                else:
                    sys.stdout = file_pos # Change the standard output to the file we created.
                    print('{1}{0}{2}'.format(os.linesep, header, positives))
                    sys.stdout = file_neg # Reset the standard output to its original value
                    print('{1}{0}{2}'.format(os.linesep, header, negatives))
                
              #  print("{1}{0}{2}".format(os.linesep, header, negatives), file=file_neg)
              #  print("{1}{0}{2}".format(os.linesep, header, positives), file=file_pos)
                
        #    print("complete")         
                
    except IOError:
        sys.exit("Error cannot open {0}".format(fasta_file))

def repeat_starting_point(str) :     
    for i in range(0, len(str)) :         
        if (str[i].islower()) :
            return i
             
    return 0

def repeat_ending_point(str) :
    """
    returns the index of the first uppercase letter after the repeat
    repeat_start_found=1 indicates that a repeat has already been found 
    """
    repeat_start_found = 0
    for i in range(0, len(str)) :  
        if (str[i].islower()):
            repeat_start_found = 1
        if (str[i].isupper()) and repeat_start_found==1 :
            return i 
             
    return 0


def fill(text, width=80):
    
    """Split text in lines of length width (=window)"""
    print(f"Splitting text in sequences of length {width}")
    positive  = ""
    negative = ""
    location = []
    random_loc = 20
    initial_gap = 0
    
    for i in range(0,14201600, width):#int(len(text)/10), width):
        #print(i)
        if text[i:i+width].isupper():
            seq = text[i:i+width] 
            if seq != "N" * width:
                negative += seq

        elif not text[i:i+width].islower(): 
   
            init_loc = repeat_starting_point(text[i:i+width]) #initial location of repeat starting point    
         #   if not init_loc == 0:
            
            
            loc = initial_gap + random.randrange(0,random_loc) #random location of repeat starting point 
            if text[(i - loc + init_loc):(i + init_loc-1)].isupper():
                
                end_loc = repeat_ending_point(text[(i - loc + init_loc):(i - loc + init_loc + width)])
                #print("ending", end_loc)
                if text[(i - loc + init_loc + end_loc+1):(i - loc + init_loc + width)].isupper():
                    positive += text[(i - loc + init_loc):(i - loc + init_loc + width)]
      
    return os.linesep.join(negative[i:i+width] for i in range(0, len(negative), width)), os.linesep.join(positive[i:i+width] for i in range(0, len(positive), width))

def fill_disease(text):
    
    """Split text in lines of length width (=window)"""
    width=len(text)
    print(f"Disease repeats: sequence of length {width}")
    
    return text[0:width]

def fill_sliding(text, width=80):
    """Split text in lines of length width (=window)"""
    print(f"Splitting text in sequences of length {width}")
    positive  = ""
    negative = ""
    location = []
    random_loc = 20 
    initial_gap = 0
    
    for i in range(0, len(text), 10):
      #  print(i)
        seq = text[i:i+width] 
        if text[i:i+width].isupper():
            
            if seq != "N" * width:
                negative += seq

        else: 
   
            positive += seq
            """
            else:
                init_ending_loc = repeat_ending_point(text[i:i+width]) #initial location of repeat ending point  
                loc = width - random_loc - random.randrange(1,random_loc) #random location of repeat ending point 
                positive += text[(i - loc + init_ending_loc):(i - loc + init_ending_loc + width)]
            """    
                
    return os.linesep.join(negative[i:i+width] for i in range(0, len(negative), width)), os.linesep.join(positive[i:i+width] for i in range(0, len(positive), width))

def fill_before(text, width=60):
    """Split text in lines of length width (=window)"""
    print(f"Splitting text in sequences of length {width}")
    positive  = ""
    negative = ""
    location = []
    random_loc = 20 
    initial_gap = 0
    gap = 10
    for i in range(0, len(text), width):
        #print(i)
        if text[i:i+width].isupper():
            seq = text[i:i+width] 
            if seq != "N" * width and text[i+width:i+2*width].isupper() and text[i-width:i].isupper():
                negative += seq

        else:
            init_loc = repeat_starting_point(text[i:i+width]) #initial location of repeat starting point    
            
            seq = text[(i + init_loc - width - gap):(i + init_loc - gap)] #the bp exactly before the repeat
            if seq.isupper():

                positive += seq
                
    return os.linesep.join(negative[i:i+width] for i in range(0, len(negative), width)), os.linesep.join(positive[i:i+width] for i in range(0, len(positive), width))

                
def fill_before_spikes(text, width=40):
    """Split text in lines of length width (=window)"""
    print(f"Splitting text in sequences of length {width}")
    positive  = ""
    negative = ""
    location = []
    random_loc = 20 
    initial_gap = 0
    starting_loc = 0 
    for i in range(0, len(text), 3*width):
        print(i)
        if text[i:i+3*width].isupper():
            seq = text[i:i+3*width] 
            if seq != "N" * 3*width:
                negative += seq

        else:
   
            init_loc = repeat_starting_point(text[i:i+3*width]) #initial location of repeat starting point 
            end_loc = repeat_ending_point(text[i:i+3*width]) #initial location of repeat ending point 
        
            seq_before = text[(i + init_loc - starting_loc -  width):(i + init_loc - starting_loc)]
            seq_pre_before = text[(i + init_loc - starting_loc -  2*width):(i + init_loc - starting_loc - width)]
            seq = text[(i + init_loc - starting_loc -  2*width):(i + init_loc - starting_loc + width)]
            if seq_before.isupper() and seq_pre_before.isupper():
                positive += seq 
                
    return os.linesep.join(negative[i:i+width] for i in range(0, len(negative), width)), os.linesep.join(positive[i:i+width] for i in range(0, len(positive), width))

def fill_spikes(text, width=80):
    """Split text in lines of length width (=window)
        repeat starts always in posistion 30
        before the 120 bp window containing the repeat there is a 120 bp that does not contain any repeat 
        after the 120 bp window containing the repeat there is again 120 bp that does not contain any repeat 
        
        negatives: 3*width bp (adjacent) with no repeats in them
        positives: 3*width bp (adjacent) with a repeat starting on the position 120+30=150 (rest of the seq before and after the repeat does not
contain any repeats    
    """
    print(f"Splitting text in sequences of length {width}")
    positive  = ""
    negative = ""
    location = []
    starting_loc = 20 
    
    for i in range(0, len(text), 3*width):
        print(i)
        if text[i:i+3*width].isupper():
            seq = text[i:i+3*width] 
            if seq != "N" * 3*width:
                negative += seq

        elif not text[i:i+3*width].islower(): 
   
            init_loc = repeat_starting_point(text[i:i+3*width]) #initial location of repeat starting point 
            end_loc = repeat_ending_point(text[i:i+3*width]) #initial location of repeat ending point 
            
            if text[(i + init_loc - starting_loc -  width):(i + init_loc - starting_loc)].isupper() and text[(i + init_loc - starting_loc + width):(i + init_loc - starting_loc + 2*width)].isupper():
                
                positive += text[(i + init_loc - starting_loc -  width):(i + init_loc - starting_loc +  2*width)]
               
                
    return os.linesep.join(negative[i:i+3*width] for i in range(0, len(negative), 3*width)), os.linesep.join(positive[i:i+3*width] for i in range(0, len(positive), 3*width))

if __name__ == '__main__':
    
    args, parser = get_parameters()
    print("Start reading {0}".format(args.fasta_file))
    split(args.fasta_file, args.spikes, args.output_dir,)
